from m2cgen.assemblers import fallback_expressions
from m2cgen.ast import BinNumExpr, CompExpr, IfExpr
from m2cgen.interpreters.utils import CachedResult, _get_handler_name


class BaseInterpreter:

    infix_expressions = (BinNumExpr, CompExpr)

    """
    Base class of AST interpreter. Provides single public method .interpret()
    which takes instance of AST expression and recursively applies method
    _do_interpret() to it.
    """
    def __init__(self):
        self._cached_expr_results = {}

    def interpret(self, expr):
        self._reset_reused_expr_cache()
        return self._do_interpret(expr)

    # Private methods implementing Visitor pattern

    def _pre_interpret_hook(self, expr, **kwargs):
        return None, kwargs

    def _do_interpret(self, expr, to_reuse=None, left_precedence=None, right_precedence=None, **kwargs):
        # Hook which allows to override kwargs and to return custom result.
        result, kwargs = self._pre_interpret_hook(expr, **kwargs)

        # If result is empty, it means that we still need to process expr.
        if result is not None:
            return result

        if expr in self._cached_expr_results:
            return self._cached_expr_results[expr].var_name

        handler = self._select_handler(expr)

        # Reset precedence on non-infix expressions.
        if not isinstance(expr, self.infix_expressions):
            left_precedence = None
            right_precedence = None

        result = handler(
            expr,
            left_precedence=left_precedence,
            right_precedence=right_precedence,
            **kwargs)

        # Note that the reuse flag passed in the arguments has a higher
        # precedence than one specified in the expression. One use case for
        # this behavior is to override the original to_reuse flag for
        # expressions that are wrapped by subroutine expression in case when
        # subroutines are not supported by specific interpreter implementation.
        expr_to_reuse = to_reuse if to_reuse is not None else expr.to_reuse
        if not expr_to_reuse:
            return result

        return self._cache_reused_expr(expr, result)

    def _cache_reused_expr(self, expr, expr_result):
        # No caching by default.
        return expr_result

    def _reset_reused_expr_cache(self):
        self._cached_expr_results = {}

    def _select_handler(self, expr):
        handler_name = _get_handler_name(type(expr))
        handler = getattr(self, handler_name, None)
        if handler is None:
            raise NotImplementedError(f"No handler found for '{type(expr).__name__}'")
        return handler


class BaseToCodeInterpreter(BaseInterpreter):

    def __init__(self, cg, feature_array_name="input"):
        super().__init__()
        self._cg = cg
        self._feature_array_name = feature_array_name


class ToCodeInterpreter(BaseToCodeInterpreter):
    """
    This interpreter provides default implementation for the methods
    interpreting AST expression into code.

    It can be used for the most programming languages and requires only
    language-specific instance of the CodeGenerator.

    !!IMPORTANT!!: Code generators used by this interpreter must know nothing
    about AST.
    """

    abs_function_name = NotImplemented
    atan_function_name = NotImplemented
    exponent_function_name = NotImplemented
    logarithm_function_name = NotImplemented
    log1p_function_name = NotImplemented
    sigmoid_function_name = NotImplemented
    softmax_function_name = NotImplemented
    sqrt_function_name = NotImplemented
    tanh_function_name = NotImplemented

    def __init__(self, cg, feature_array_name="input"):
        super().__init__(cg, feature_array_name=feature_array_name)
        self.with_vectors = False
        self.with_math_module = False

    def interpret_id_expr(self, expr, **kwargs):
        return self._do_interpret(expr.expr, **kwargs)

    def interpret_comp_expr(self, expr, **kwargs):
        op = self._cg._comp_op_overwrite(expr.op)
        return self._cg.infix_expression(
            left=self._do_interpret(expr.left, **kwargs),
            op=op,
            right=self._do_interpret(expr.right, **kwargs))

    def interpret_bin_num_expr(self, expr, left_precedence=None, right_precedence=None, **kwargs):
        return self._cg.infix_expression(
            left=self._do_interpret(expr.left, left_precedence=expr.precedence, **kwargs),
            op=expr.op.value,
            right=self._do_interpret(expr.right, right_precedence=expr.precedence, **kwargs),
            wrap=self._wrap_infix_expr(expr, left_precedence, right_precedence))

    def interpret_num_val(self, expr, **kwargs):
        return self._cg.num_value(value=expr.value)

    def interpret_feature_ref(self, expr, **kwargs):
        return self._cg.array_index_access(
            array_name=self._feature_array_name,
            index=expr.index)

    def interpret_vector_val(self, expr, **kwargs):
        self.with_vectors = True
        nested = [self._do_interpret(expr, **kwargs) for expr in expr.exprs]
        return self._cg.vector_init(nested)

    def interpret_abs_expr(self, expr, **kwargs):
        if self.abs_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.abs(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.abs_function_name, nested_result)

    def interpret_atan_expr(self, expr, **kwargs):
        if self.atan_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.atan(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.atan_function_name, nested_result)

    def interpret_exp_expr(self, expr, **kwargs):
        if self.exponent_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.exp(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.exponent_function_name, nested_result)

    def interpret_log_expr(self, expr, **kwargs):
        if self.logarithm_function_name is NotImplemented:
            raise NotImplementedError("Logarithm function is not provided")
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.logarithm_function_name, nested_result)

    def interpret_log1p_expr(self, expr, **kwargs):
        if self.log1p_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.log1p(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.log1p_function_name, nested_result)

    def interpret_sigmoid_expr(self, expr, **kwargs):
        if self.sigmoid_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.sigmoid(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.sigmoid_function_name, nested_result)

    def interpret_softmax_expr(self, expr, **kwargs):
        if self.softmax_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.softmax(expr.exprs), **kwargs)
        self.with_vectors = True
        self.with_math_module = True
        nested = [self._do_interpret(expr, **kwargs) for expr in expr.exprs]
        return self._cg.function_invocation(self.softmax_function_name, self._cg.vector_init(nested))

    def interpret_sqrt_expr(self, expr, **kwargs):
        if self.sqrt_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.sqrt(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.sqrt_function_name, nested_result)

    def interpret_tanh_expr(self, expr, **kwargs):
        if self.tanh_function_name is NotImplemented:
            return self._do_interpret(fallback_expressions.tanh(expr.expr), **kwargs)
        self.with_math_module = True
        nested_result = self._do_interpret(expr.expr, **kwargs)
        return self._cg.function_invocation(self.tanh_function_name, nested_result)

    def _wrap_infix_expr(self, expr, left_precedence, right_precedence):
        wrap = left_precedence is not None and left_precedence > expr.precedence
        wrap |= right_precedence is not None and right_precedence >= expr.precedence
        return wrap


class ImperativeToCodeInterpreter(ToCodeInterpreter):
    """
    This interpreter provides default implementation for the methods
    interpreting AST expression into code.

    It can be used for the most programming languages and requires only
    language-specific instance of the CodeGenerator.

    !!IMPORTANT!!: Code generators used by this interpreter must know nothing
    about AST.
    """

    def interpret_if_expr(self, expr, if_var_name=None, **kwargs):
        if if_var_name is not None:
            var_name = if_var_name
        else:
            var_name = self._cg.add_var_declaration(expr.output_size)

        def handle_nested_expr(nested):
            if isinstance(nested, IfExpr):
                self._do_interpret(nested, if_var_name=var_name, **kwargs)
            else:
                nested_result = self._do_interpret(nested, **kwargs)
                self._cg.add_var_assignment(var_name, nested_result, nested.output_size)

        self._cg.add_if_statement(self._do_interpret(expr.test, **kwargs))
        handle_nested_expr(expr.body)
        self._cg.add_else_statement()
        handle_nested_expr(expr.orelse)
        self._cg.add_block_termination()

        return var_name

    def _cache_reused_expr(self, expr, expr_result):
        var_name = self._cg.add_var_declaration(expr.output_size)
        self._cg.add_var_assignment(var_name, expr_result, expr.output_size)
        self._cached_expr_results[expr] = CachedResult(var_name=var_name, expr_result=None)
        return var_name


class FunctionalToCodeInterpreter(ToCodeInterpreter):
    """
    This interpreter provides default implementation for the methods
    interpreting AST expression into code.

    It can be used for the most functional programming languages and requires
    only language-specific instance of the CodeGenerator.

    !!IMPORTANT!!: Code generators used by this interpreter must know nothing
    about AST.
    """

    def interpret_if_expr(self, expr, if_code_gen=None, **kwargs):
        if if_code_gen is None:
            code_gen = self.create_code_generator()
            nested = False
        else:
            code_gen = if_code_gen
            nested = True

        code_gen.add_if_statement(self._do_interpret(expr.test, **kwargs))
        code_gen.add_code_line(self._do_interpret(expr.body, if_code_gen=code_gen, **kwargs))
        code_gen.add_else_statement()
        code_gen.add_code_line(self._do_interpret(expr.orelse, if_code_gen=code_gen, **kwargs))
        code_gen.add_if_termination()

        if not nested:
            return self._cache_reused_expr(expr, code_gen.finalize_and_get_generated_code())

    # Cached expressions become functions with no arguments, i.e. values
    # which are CAFs. Therefore, they are computed only once.
    def _cache_reused_expr(self, expr, expr_result):
        if expr in self._cached_expr_results:
            return self._cached_expr_results[expr].var_name
        else:
            func_name = self._cg.get_func_name()
            self._cached_expr_results[expr] = CachedResult(var_name=func_name, expr_result=expr_result)
            return func_name

    def create_code_generator(self):
        raise NotImplementedError



from m2cgen import assemblers, ast


class RandomForestModelAssembler(assemblers.base.ModelAssembler):

    def assemble(self):
        trees = self.model.estimators_

        def assemble_tree_expr(t):
            assembler = assemblers.TreeModelAssembler(t)

            return assembler.assemble()

        assembled_trees = [assemble_tree_expr(t) for t in trees]
        return assemblers.utils.apply_bin_op(
            assemblers.utils.apply_op_to_expressions(ast.BinNumOpType.ADD, *assembled_trees),
            ast.NumVal(1 / self.model.n_estimators),
            ast.BinNumOpType.MUL)



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CodeTemplate, ImperativeCodeGenerator


class PythonCodeGenerator(ImperativeCodeGenerator):

    tpl_num_value = CodeTemplate("{value}")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_return_statement = CodeTemplate("return {value}")
    tpl_array_index_access = CodeTemplate("{array_name}[{index}]")
    tpl_if_statement = CodeTemplate("if {if_def}:")
    tpl_else_statement = CodeTemplate("else:")
    tpl_var_assignment = CodeTemplate("{var_name} = {value}")

    tpl_var_declaration = CodeTemplate("")
    tpl_block_termination = CodeTemplate("")

    def add_function_def(self, name, args):
        function_def = f"def {name}({', '.join(args)}):"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield

    def vector_init(self, values):
        return f"[{', '.join(values)}]"

    def add_dependency(self, dep):
        self.prepend_code_line(f"import {dep}")



"""This module provides an implementation for a variety of functions
expressed in library's AST.

These AST-based implementations are used as fallbacks in case
when the target language lacks native support for respective functions
provided in this module.
"""
import math

from m2cgen import ast
from m2cgen.assemblers import utils


def abs(expr):
    expr = ast.IdExpr(expr, to_reuse=True)
    return ast.IfExpr(
        utils.lt(expr, ast.NumVal(0.0)),
        utils.sub(ast.NumVal(0.0), expr),
        expr)


def tanh(expr):
    expr = ast.IdExpr(expr, to_reuse=True)
    tanh_expr = utils.sub(
        ast.NumVal(1.0),
        utils.div(
            ast.NumVal(2.0),
            utils.add(
                ast.ExpExpr(
                    utils.mul(
                        ast.NumVal(2.0),
                        expr)),
                ast.NumVal(1.0))))
    return ast.IfExpr(
        utils.gt(expr, ast.NumVal(44.0)),  # exp(2*x) <= 2^127
        ast.NumVal(1.0),
        ast.IfExpr(
            utils.lt(expr, ast.NumVal(-44.0)),
            ast.NumVal(-1.0),
            tanh_expr))


def sqrt(expr):
    return ast.PowExpr(
        base_expr=expr,
        exp_expr=ast.NumVal(0.5))


def exp(expr):
    return ast.PowExpr(
        base_expr=ast.NumVal(math.e),
        exp_expr=expr)


def log1p(expr):
    # Use trick to compute log1p for small values more accurate
    # https://www.johndcook.com/blog/2012/07/25/trick-for-computing-log1x/
    expr = ast.IdExpr(expr, to_reuse=True)
    expr1p = utils.add(ast.NumVal(1.0), expr, to_reuse=True)
    expr1pm1 = utils.sub(expr1p, ast.NumVal(1.0), to_reuse=True)
    return ast.IfExpr(
        utils.eq(expr1pm1, ast.NumVal(0.0)),
        expr,
        utils.div(utils.mul(expr, ast.LogExpr(expr1p)), expr1pm1))


def atan(expr):
    expr = ast.IdExpr(expr, to_reuse=True)
    expr_abs = ast.AbsExpr(expr, to_reuse=True)

    expr_reduced = ast.IdExpr(
        ast.IfExpr(
            utils.gt(expr_abs, ast.NumVal(2.4142135623730950488)),
            utils.div(ast.NumVal(1.0), expr_abs),
            ast.IfExpr(
                utils.gt(expr_abs, ast.NumVal(0.66)),
                utils.div(
                    utils.sub(expr_abs, ast.NumVal(1.0)),
                    utils.add(expr_abs, ast.NumVal(1.0))),
                expr_abs)),
        to_reuse=True)

    P0 = ast.NumVal(-8.750608600031904122785e-01)
    P1 = ast.NumVal(1.615753718733365076637e+01)
    P2 = ast.NumVal(7.500855792314704667340e+01)
    P3 = ast.NumVal(1.228866684490136173410e+02)
    P4 = ast.NumVal(6.485021904942025371773e+01)
    Q0 = ast.NumVal(2.485846490142306297962e+01)
    Q1 = ast.NumVal(1.650270098316988542046e+02)
    Q2 = ast.NumVal(4.328810604912902668951e+02)
    Q3 = ast.NumVal(4.853903996359136964868e+02)
    Q4 = ast.NumVal(1.945506571482613964425e+02)
    expr2 = utils.mul(expr_reduced, expr_reduced, to_reuse=True)
    z = utils.mul(
        expr2,
        utils.div(
            utils.sub(
                utils.mul(
                    expr2,
                    utils.sub(
                        utils.mul(
                            expr2,
                            utils.sub(
                                utils.mul(
                                    expr2,
                                    utils.sub(
                                        utils.mul(
                                            expr2,
                                            P0
                                        ),
                                        P1
                                    )
                                ),
                                P2
                            )
                        ),
                        P3
                    )
                ),
                P4
            ),
            utils.add(
                Q4,
                utils.mul(
                    expr2,
                    utils.add(
                        Q3,
                        utils.mul(
                            expr2,
                            utils.add(
                                Q2,
                                utils.mul(
                                    expr2,
                                    utils.add(
                                        Q1,
                                        utils.mul(
                                            expr2,
                                            utils.add(
                                                Q0,
                                                expr2
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )
    )
    z = utils.add(utils.mul(expr_reduced, z), expr_reduced)

    ret = utils.mul(
        z,
        ast.IfExpr(
            utils.gt(expr_abs, ast.NumVal(2.4142135623730950488)),
            ast.NumVal(-1.0),
            ast.NumVal(1.0)))
    ret = utils.add(
        ret,
        ast.IfExpr(
            utils.lte(expr_abs, ast.NumVal(0.66)),
            ast.NumVal(0.0),
            ast.IfExpr(
                utils.gt(expr_abs, ast.NumVal(2.4142135623730950488)),
                ast.NumVal(1.570796326794896680463661649),
                ast.NumVal(0.7853981633974483402318308245))))
    ret = utils.mul(
        ret,
        ast.IfExpr(
            utils.lt(expr, ast.NumVal(0.0)),
            ast.NumVal(-1.0),
            ast.NumVal(1.0)))

    return ret


def sigmoid(expr, to_reuse=False):
    neg_expr = ast.BinNumExpr(ast.NumVal(0.0), expr, ast.BinNumOpType.SUB)
    exp_expr = ast.ExpExpr(neg_expr)
    return ast.BinNumExpr(
        ast.NumVal(1.0),
        ast.BinNumExpr(ast.NumVal(1.0), exp_expr, ast.BinNumOpType.ADD),
        ast.BinNumOpType.DIV,
        to_reuse=to_reuse)


def softmax(exprs):
    exp_exprs = [ast.ExpExpr(e, to_reuse=True) for e in exprs]
    exp_sum_expr = utils.apply_op_to_expressions(ast.BinNumOpType.ADD, *exp_exprs, to_reuse=True)
    return ast.VectorVal([
        ast.BinNumExpr(e, exp_sum_expr, ast.BinNumOpType.DIV)
        for e in exp_exprs
    ])



from contextlib import contextmanager

from m2cgen.ast import CompOpType
from m2cgen.interpreters.code_generator import CodeTemplate, FunctionalCodeGenerator


class HaskellCodeGenerator(FunctionalCodeGenerator):
    tpl_function_signature = CodeTemplate("{function_name} =")
    tpl_if_statement = CodeTemplate("if {if_def} then")
    tpl_else_statement = CodeTemplate("else")
    tpl_num_value = CodeTemplate("{value}")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_module_definition = CodeTemplate("module {module_name} where")

    def num_value(self, value):
        result = super().num_value(value)
        return f"({result})" if result[:1] == "-" else result

    def array_index_access(self, array_name, index):
        return self.tpl_infix_expression(
            left=array_name, op="!!", right=index)

    def add_if_termination(self):
        self.decrease_indent()

    def add_function_def(self, name, args, is_scalar_output):
        types = " -> ".join(
            ["[Double]" if is_vector else "Double"
             for is_vector, _ in [*args, (not is_scalar_output, None)]])
        signature = f"{name} :: {types}"
        self.add_code_line(signature)

        func_args = " ".join([n for _, n in args])
        function_def = f"{name} {func_args} ="
        self.add_code_line(function_def)

        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args, is_scalar_output):
        self.add_function_def(name, args, is_scalar_output)
        yield
        self.decrease_indent()

    def vector_init(self, values):
        return f"[{', '.join(values)}]"

    def _comp_op_overwrite(self, op):
        if op == CompOpType.NOT_EQ:
            return "/="
        else:
            return op.value



"""CLI for m2cgen.

Example usage:
    $ m2cgen <path_to_file> --language java --class_name MyModel --package_name foo.bar.baz
    $ m2cgen --language java < <path_to_file>

Model can also be piped:
    # cat <path_to_file> | m2cgen --language java
"""
import pickle
import sys
from argparse import ArgumentParser, FileType
from inspect import signature

import numpy as np

import m2cgen

LANGUAGE_TO_EXPORTER = {
    "python": (m2cgen.export_to_python, ["indent", "function_name"]),
    "java": (m2cgen.export_to_java, ["indent", "class_name", "package_name", "function_name"]),
    "c": (m2cgen.export_to_c, ["indent", "function_name"]),
    "go": (m2cgen.export_to_go, ["indent", "function_name"]),
    "javascript": (m2cgen.export_to_javascript, ["indent", "function_name"]),
    "visual_basic": (m2cgen.export_to_visual_basic, ["module_name", "indent", "function_name"]),
    "c_sharp": (m2cgen.export_to_c_sharp, ["indent", "class_name", "namespace", "function_name"]),
    "powershell": (m2cgen.export_to_powershell, ["indent", "function_name"]),
    "r": (m2cgen.export_to_r, ["indent", "function_name"]),
    "php": (m2cgen.export_to_php, ["indent", "function_name"]),
    "dart": (m2cgen.export_to_dart, ["indent", "function_name"]),
    "haskell": (m2cgen.export_to_haskell, ["module_name", "indent", "function_name"]),
    "ruby": (m2cgen.export_to_ruby, ["indent", "function_name"]),
    "f_sharp": (m2cgen.export_to_f_sharp, ["indent", "function_name"]),
    "rust": (m2cgen.export_to_rust, ["indent", "function_name"]),
    "elixir": (m2cgen.export_to_elixir, ["module_name", "indent", "function_name"]),
}


# The maximum recursion depth is represented by the maximum int32 value.
MAX_RECURSION_DEPTH = np.iinfo(np.intc).max


parser = ArgumentParser(
    prog="m2cgen",
    description="Generate code in native language for provided model.")
parser.add_argument(
    "infile",
    type=FileType("rb"),
    nargs="?",
    default=sys.stdin.buffer,
    help="File with pickle representation of the model.")
parser.add_argument(
    "--language", "-l",
    type=str,
    choices=LANGUAGE_TO_EXPORTER.keys(),
    help="Target language.",
    required=True)
parser.add_argument(
    "--function_name", "-fn",
    dest="function_name",
    type=str,
    # The default value is conditional and will be set in the argument's
    # post-processing, based on the signature of the `export` function
    # that belongs to the specified target language.
    default=None,
    help="Name of the function in the generated code.")
parser.add_argument(
    "--class_name", "-cn",
    dest="class_name",
    type=str,
    help="Name of the generated class (if supported by target language).")
parser.add_argument(
    "--package_name", "-pn",
    dest="package_name",
    type=str,
    help="Package name for the generated code (if supported by target language).")
parser.add_argument(
    "--module_name", "-mn",
    dest="module_name",
    type=str,
    help="Module name for the generated code (if supported by target language).")
parser.add_argument(
    "--namespace", "-ns",
    dest="namespace",
    type=str,
    help="Namespace for the generated code (if supported by target language).")
parser.add_argument(
    "--indent", "-i",
    dest="indent",
    type=int,
    default=4,
    help="Indentation for the generated code.")
parser.add_argument(
    "--recursion-limit", "-rl",
    type=int,
    help="Sets the maximum depth of the Python interpreter stack. No limit by default",
    default=MAX_RECURSION_DEPTH)
parser.add_argument(
    "--version", "-v",
    action="version",
    version=f"%(prog)s {m2cgen.__version__}")


def parse_args(args):
    return parser.parse_args(args)


def generate_code(args):
    sys.setrecursionlimit(args.recursion_limit)

    with args.infile as f:
        model = pickle.load(f)

    exporter, supported_args = LANGUAGE_TO_EXPORTER[args.language]

    kwargs = {}
    for arg_name in supported_args:
        arg_value = getattr(args, arg_name)
        if arg_value is not None:
            kwargs[arg_name] = arg_value

        # Special handling for the function_name parameter, which needs to be
        # the same as the default value of the keyword argument of the exporter
        # (this is due to languages like C# which prefer their method names to
        # follow PascalCase unlike all the other supported languages -- see
        # https://github.com/BayesWitnesses/m2cgen/pull/166#discussion_r379867601
        # for more).
        if arg_name == 'function_name' and arg_value is None:
            param = signature(exporter).parameters['function_name']
            kwargs[arg_name] = param.default

    return exporter(model, **kwargs)


def main():
    args = parse_args(sys.argv[1:])
    print(generate_code(args))



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CLikeCodeGenerator


class DartCodeGenerator(CLikeCodeGenerator):

    scalar_type = "double"
    vector_type = "List<double>"

    def add_function_def(self, name, args, is_vector_output):
        return_type = self._get_var_declare_type(is_vector_output)
        func_args = ",".join([
            f"{self._get_var_declare_type(is_vector)} {n}"
            for is_vector, n in args])
        function_def = f"{return_type} {name}({func_args}) {{"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args, is_vector_output):
        self.add_function_def(name, args, is_vector_output)
        yield
        self.add_block_termination()

    def method_invocation(self, method_name, obj, args):
        return f"({obj}).{method_name}({', '.join(map(str, args))})"

    def vector_init(self, values):
        return f"[{', '.join(values)}]"

    def _get_var_declare_type(self, is_vector):
        return self.vector_type if is_vector else self.scalar_type

    def add_dependency(self, dep):
        self.prepend_code_line(f"import '{dep}';")



from pathlib import Path

from m2cgen.ast import BinNumOpType
from m2cgen.interpreters.elixir.code_generator import ElixirCodeGenerator
from m2cgen.interpreters.interpreter import FunctionalToCodeInterpreter
from m2cgen.interpreters.mixins import BinExpressionDepthTrackingMixin, LinearAlgebraMixin, PowExprFunctionMixin
from m2cgen.interpreters.utils import get_file_content


class ElixirInterpreter(FunctionalToCodeInterpreter,
                        PowExprFunctionMixin,
                        LinearAlgebraMixin,
                        BinExpressionDepthTrackingMixin):

    bin_depth_threshold = 15

    supported_bin_vector_ops = {
        BinNumOpType.ADD: "add_vectors",
    }

    supported_bin_vector_num_ops = {
        BinNumOpType.MUL: "mul_vector_number",
    }

    abs_function_name = "abs"
    atan_function_name = ":math.atan"
    exponent_function_name = ":math.exp"
    logarithm_function_name = ":math.log"
    log1p_function_name = "log1p"
    sigmoid_function_name = "sigmoid"
    softmax_function_name = "softmax"
    sqrt_function_name = ":math.sqrt"
    tanh_function_name = ":math.tanh"
    power_function_name = ":math.pow"

    with_log1p_expr = False
    with_sigmoid_expr = False
    with_softmax_expr = False

    def __init__(self,  module_name="Model", indent=4, function_name="score",
                 *args, **kwargs):
        self.module_name = module_name
        self.indent = indent
        self.function_name = function_name

        super().__init__(self.create_code_generator(), *args, **kwargs)

    def interpret(self, expr):
        self._cg.reset_state()
        self._reset_reused_expr_cache()

        args = [(True, self._feature_array_name)]
        func_name = self.function_name

        with self._cg.function_definition(
                name=func_name,
                args=args,
                is_scalar_output=expr.output_size == 1):
            last_result = self._do_interpret(expr)
            self._dump_cache()
            self._cg.add_code_line(last_result)

        current_dir = Path(__file__).absolute().parent

        if self.with_linear_algebra:
            filename = current_dir / "linear_algebra.ex"
            self._cg.add_code_lines(get_file_content(filename))

        if self.with_log1p_expr:
            filename = current_dir / "log1p.ex"
            self._cg.add_code_lines(get_file_content(filename))

        if self.with_softmax_expr:
            filename = current_dir / "softmax.ex"
            self._cg.add_code_lines(get_file_content(filename))

        if self.with_sigmoid_expr:
            filename = current_dir / "sigmoid.ex"
            self._cg.add_code_lines(get_file_content(filename))

        self._cg.prepend_code_line(self._cg.tpl_module_definition(
            module_name=self.module_name))

        self._cg.add_code_line(self._cg.tpl_block_termination())

        return self._cg.finalize_and_get_generated_code()

    def create_code_generator(self):
        return ElixirCodeGenerator(indent=self.indent)

    def interpret_log1p_expr(self, expr, **kwargs):
        self.with_log1p_expr = True
        return super().interpret_log1p_expr(expr, **kwargs)

    def interpret_softmax_expr(self, expr, **kwargs):
        self.with_softmax_expr = True
        return super().interpret_softmax_expr(expr, **kwargs)

    def interpret_sigmoid_expr(self, expr, **kwargs):
        self.with_sigmoid_expr = True
        return super().interpret_sigmoid_expr(expr, **kwargs)

    def _dump_cache(self):
        if self._cached_expr_results:
            for func_name, expr_result in self._cached_expr_results.values():
                self._cg.add_function(
                    function_name=func_name, function_body=expr_result)

    def bin_depth_threshold_hook(self, expr, **kwargs):
        if expr in self._cached_expr_results:
            return self._cached_expr_results[expr].var_name
        result = self._do_interpret(expr, **kwargs)
        return self._cache_reused_expr(expr, result)



from contextlib import contextmanager

from m2cgen.ast import CompOpType
from m2cgen.interpreters.code_generator import CLikeCodeGenerator, CodeTemplate


class PhpCodeGenerator(CLikeCodeGenerator):

    tpl_array_index_access = CodeTemplate("${array_name}[{index}]")

    def add_function_def(self, name, args):
        func_args = ", ".join([
            f"{'array ' if is_vector else ''}${n}"
            for is_vector, n in args])
        function_def = f"function {name}({func_args}) {{"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield
        self.add_block_termination()

    def get_var_name(self):
        return f"${super().get_var_name()}"

    def add_var_declaration(self, size):
        var_name = self.get_var_name()
        self.add_var_assignment(
            var_name=var_name,
            value="array()" if size > 1 else "null",
            value_size=size)
        return var_name

    def vector_init(self, values):
        return f"array({', '.join(values)})"

    def _comp_op_overwrite(self, op):
        if op == CompOpType.EQ:
            return "==="
        elif op == CompOpType.NOT_EQ:
            return "!=="
        else:
            return op.value



from io import StringIO
from weakref import finalize

import numpy as np

from m2cgen.interpreters.utils import format_float


class CodeTemplate:

    def __init__(self, template):
        self.str_template = template

    def __call__(self, *args, **kwargs):

        def _is_float(value):
            return isinstance(value, (float, np.floating))

        return self.str_template.format(
            *[format_float(i) if _is_float(i) else i for i in args],
            **{k: format_float(v) if _is_float(v) else v for k, v in kwargs.items()})


class BaseCodeGenerator:
    """
    This class provides basic functionality to generate code. It is
    language-agnostic, but exposes set of attributes which subclasses should
    use to define syntax specific for certain language(s).

    !!IMPORTANT!!: Code generators must know nothing about AST.
    """

    tpl_num_value = NotImplemented
    tpl_infix_expression = NotImplemented
    tpl_array_index_access = NotImplemented

    def __init__(self, indent=4):
        self._indent = indent
        self._code_buf = None
        self.reset_state()

    def reset_state(self):
        self._current_indent = 0
        self._finalize_buffer()
        self._code_buf = StringIO()
        self._code = None
        self._finalizer = finalize(self, self._finalize_buffer)

    def _finalize_buffer(self):
        if self._code_buf is not None and not self._code_buf.closed:
            self._code_buf.close()

    def _write_to_code_buffer(self, text, prepend=False):
        if self._code_buf.closed:
            raise BufferError(
                "Cannot modify code after getting generated code and closing the underlying buffer!\n"
                "Call reset_state() to allocate new buffer.")
        if prepend:
            self._code_buf.seek(0)
            old_content = self._code_buf.read()
            self._code_buf.seek(0)
            text += old_content
        self._code_buf.write(text)

    def finalize_and_get_generated_code(self):
        if not self._code_buf.closed:
            self._code = self._code_buf.getvalue()
            self._finalize_buffer()
        return self._code if self._code is not None else ""

    def increase_indent(self):
        self._current_indent += self._indent

    def decrease_indent(self):
        self._current_indent -= self._indent
        assert self._current_indent >= 0, f"Invalid indentation: {self._current_indent}"

    # All code modifications should be implemented via following methods.

    def add_code_line(self, line):
        if not line:
            return
        self.add_code_lines([line.strip()])

    def add_code_lines(self, lines):
        if isinstance(lines, str):
            lines = lines.strip().split("\n")
        indent = " " * self._current_indent
        self._write_to_code_buffer(indent + f"\n{indent}".join(lines) + "\n")

    def prepend_code_line(self, line):
        if not line:
            return
        self.prepend_code_lines([line.strip()])

    def prepend_code_lines(self, lines):
        new_line = "\n"
        if isinstance(lines, str):
            lines = lines.strip().split(new_line)
        self._write_to_code_buffer(f"{new_line.join(lines)}{new_line}", prepend=True)

    # Following methods simply compute expressions using templates without changing result.

    def infix_expression(self, left, right, op, wrap=False):
        result = self.tpl_infix_expression(left=left, right=right, op=op)
        return result if not wrap else f"({result})"

    def num_value(self, value):
        return self.tpl_num_value(value=value)

    def array_index_access(self, array_name, index):
        return self.tpl_array_index_access(array_name=array_name, index=index)

    def function_invocation(self, function_name, *args):
        return f"{function_name}({', '.join(map(str, args))})"

    # Helpers

    def _comp_op_overwrite(self, op):
        return op.value


class ImperativeCodeGenerator(BaseCodeGenerator):
    """
    This class provides basic functionality to generate code. It is
    language-agnostic, but exposes set of attributes which subclasses should
    use to define syntax specific for certain language(s).

    !!IMPORTANT!!: Code generators must know nothing about AST.
    """

    tpl_var_declaration = NotImplemented
    tpl_return_statement = NotImplemented
    tpl_if_statement = NotImplemented
    tpl_else_statement = NotImplemented
    tpl_block_termination = NotImplemented
    tpl_var_assignment = NotImplemented

    def reset_state(self):
        super().reset_state()
        self._var_idx = 0

    def get_var_name(self):
        var_name = f"var{self._var_idx}"
        self._var_idx += 1
        return var_name

    # Following statements compute expressions using templates AND add
    # it to the result.

    def add_return_statement(self, value):
        self.add_code_line(self.tpl_return_statement(value=value))

    def add_var_declaration(self, size):
        var_name = self.get_var_name()
        is_vector = size > 1
        var_type = self._get_var_declare_type(is_vector)
        self.add_code_line(
            self.tpl_var_declaration(
                var_type=var_type, var_name=var_name))
        return var_name

    def add_if_statement(self, if_def):
        self.add_code_line(self.tpl_if_statement(if_def=if_def))
        self.increase_indent()

    def add_else_statement(self):
        self.decrease_indent()
        self.add_code_line(self.tpl_else_statement())
        self.increase_indent()

    def add_block_termination(self):
        self.decrease_indent()
        self.add_code_line(self.tpl_block_termination())

    def add_var_assignment(self, var_name, value, value_size):
        self.add_code_line(self.tpl_var_assignment(var_name=var_name, value=value))

    # Helpers

    def _get_var_declare_type(self, expr):
        return NotImplemented


class CLikeCodeGenerator(ImperativeCodeGenerator):
    """
    This code generator provides C-like syntax so that subclasses will only
    have to provide logic for wrapping expressions into functions/classes/etc.
    """

    tpl_num_value = CodeTemplate("{value}")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_var_declaration = CodeTemplate("{var_type} {var_name};")
    tpl_return_statement = CodeTemplate("return {value};")
    tpl_array_index_access = CodeTemplate("{array_name}[{index}]")
    tpl_if_statement = CodeTemplate("if ({if_def}) {{")
    tpl_else_statement = CodeTemplate("}} else {{")
    tpl_block_termination = CodeTemplate("}}")
    tpl_var_assignment = CodeTemplate("{var_name} = {value};")


class FunctionalCodeGenerator(BaseCodeGenerator):
    """
    This class provides basic functionality to generate code. It is
    language-agnostic, but exposes set of attributes which subclasses should
    use to define syntax specific for certain language(s).

    !!IMPORTANT!!: Code generators must know nothing about AST.
    """

    tpl_function_signature = NotImplemented
    tpl_if_statement = NotImplemented
    tpl_else_statement = NotImplemented
    tpl_block_termination = NotImplemented

    def reset_state(self):
        super().reset_state()
        self._func_idx = 0

    def get_func_name(self):
        func_name = f"func{self._func_idx}"
        self._func_idx += 1
        return func_name

    # Following statements compute expressions using templates AND add it to the result.

    def add_function(self, function_name, function_body):
        self.add_code_line(self.tpl_function_signature(function_name=function_name))
        self.increase_indent()
        self.add_code_lines(function_body)
        self.decrease_indent()

    def function_invocation(self, function_name, *args):
        function_args = " ".join(map(lambda x: f"({x})", args))
        return f"{function_name} {function_args}"

    def add_if_statement(self, if_def):
        self.add_code_line(self.tpl_if_statement(if_def=if_def))
        self.increase_indent()

    def add_else_statement(self):
        self.decrease_indent()
        self.add_code_line(self.tpl_else_statement())
        self.increase_indent()



from contextlib import contextmanager

from m2cgen.ast import CompOpType
from m2cgen.interpreters.code_generator import CLikeCodeGenerator, CodeTemplate


class PowershellCodeGenerator(CLikeCodeGenerator):

    tpl_var_declaration = CodeTemplate("{var_type}{var_name} = {init_val}")
    tpl_var_assignment = CodeTemplate("{var_name} = {value}")
    tpl_array_index_access = CodeTemplate("${array_name}[{index}]")
    tpl_return_statement = CodeTemplate("return {value}")

    scalar_type = "[double]"
    vector_type = "[double[]]"

    operator_map = {
        CompOpType.EQ: "-eq",
        CompOpType.NOT_EQ: "-ne",
        CompOpType.GTE: "-ge",
        CompOpType.LTE: "-le",
        CompOpType.GT: "-gt",
        CompOpType.LT: "-lt"
    }

    def add_function_def(self, name, args):
        func_args = ", ".join([
            f"{self._get_var_declare_type(is_vector)} ${n}"
            for is_vector, n in args])
        function_def = f"function {name}({func_args}) {{"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield
        self.add_block_termination()

    def function_invocation(self, function_name, *args):
        functions_args = " ".join(map(lambda x: f"$({x})", args))
        return f"{function_name} {functions_args}"

    def math_function_invocation(self, function_name, *args):
        return f"{function_name}({', '.join(map(str, args))})"

    def get_var_name(self):
        return f"${super().get_var_name()}"

    def add_var_declaration(self, size):
        var_name = self.get_var_name()
        self.add_code_line(
            self.tpl_var_declaration(var_type=self._get_var_declare_type(size > 1),
                                     var_name=var_name,
                                     init_val="@(0.0)" if size > 1 else "0.0"))
        return var_name

    def vector_init(self, values):
        vals = ", ".join(map(lambda x: f"$({x})", values))
        return f"@({vals})"

    def _get_var_declare_type(self, is_vector):
        return self.vector_type if is_vector else self.scalar_type

    def _comp_op_overwrite(self, op):
        return self.operator_map[op]



from pathlib import Path

from m2cgen.ast import BinNumOpType
from m2cgen.interpreters.interpreter import ImperativeToCodeInterpreter
from m2cgen.interpreters.mixins import LinearAlgebraMixin, PowExprFunctionMixin
from m2cgen.interpreters.rust.code_generator import RustCodeGenerator
from m2cgen.interpreters.utils import get_file_content


class RustInterpreter(ImperativeToCodeInterpreter,
                      PowExprFunctionMixin,
                      LinearAlgebraMixin):

    supported_bin_vector_ops = {
        BinNumOpType.ADD: "add_vectors",
    }

    supported_bin_vector_num_ops = {
        BinNumOpType.MUL: "mul_vector_number",
    }

    abs_function_name = "f64::abs"
    atan_function_name = "f64::atan"
    exponent_function_name = "f64::exp"
    logarithm_function_name = "f64::ln"
    log1p_function_name = "f64::ln_1p"
    power_function_name = "f64::powf"
    sigmoid_function_name = "sigmoid"
    softmax_function_name = "softmax"
    sqrt_function_name = "f64::sqrt"
    tanh_function_name = "f64::tanh"

    with_sigmoid_expr = False
    with_softmax_expr = False

    def __init__(self, indent=4, function_name="score", *args, **kwargs):
        self.function_name = function_name

        cg = RustCodeGenerator(indent=indent)
        super().__init__(cg, *args, **kwargs)

    def interpret(self, expr):
        self._cg.reset_state()
        self._reset_reused_expr_cache()

        with self._cg.function_definition(
                name=self.function_name,
                args=[(True, self._feature_array_name)],
                is_scalar_output=expr.output_size == 1):
            last_result = self._do_interpret(expr)
            self._cg.add_return_statement(last_result)

        current_dir = Path(__file__).absolute().parent

        if self.with_linear_algebra:
            filename = current_dir / "linear_algebra.rs"
            self._cg.add_code_lines(get_file_content(filename))

        if self.with_softmax_expr:
            filename = current_dir / "softmax.rs"
            self._cg.add_code_lines(get_file_content(filename))

        if self.with_sigmoid_expr:
            filename = current_dir / "sigmoid.rs"
            self._cg.add_code_lines(get_file_content(filename))

        return self._cg.finalize_and_get_generated_code()

    def interpret_sigmoid_expr(self, expr, **kwargs):
        self.with_sigmoid_expr = True
        return super().interpret_sigmoid_expr(expr, **kwargs)

    def interpret_softmax_expr(self, expr, **kwargs):
        self.with_softmax_expr = True
        return super().interpret_softmax_expr(expr, **kwargs)



from pathlib import Path

from m2cgen.interpreters.interpreter import ImperativeToCodeInterpreter
from m2cgen.interpreters.mixins import (
    BinExpressionDepthTrackingMixin,
    LinearAlgebraMixin,
    PowExprInfixMixin,
    SubroutinesMixin
)
from m2cgen.interpreters.r.code_generator import RCodeGenerator
from m2cgen.interpreters.utils import get_file_content


class RInterpreter(ImperativeToCodeInterpreter,
                   PowExprInfixMixin,
                   LinearAlgebraMixin,
                   BinExpressionDepthTrackingMixin,
                   SubroutinesMixin):

    # R doesn't allow to have more than 50 nested if, [, [[, {, ( calls.
    # It raises contextstack overflow error not only for explicitly nested
    # calls, but also if met above mentioned number of parentheses
    # in one expression. Given that there is no way to control
    # the number of parentheses in one expression for now,
    # the following variable set to 50 / 2 value is expected to prevent
    # contextstack overflow error occurrence.
    # This value is just a heuristic and is subject to change in the future
    # based on the users' feedback.
    bin_depth_threshold = 25

    ast_size_check_frequency = 2
    ast_size_per_subroutine_threshold = 200

    abs_function_name = "abs"
    atan_function_name = "atan"
    exponent_function_name = "exp"
    logarithm_function_name = "log"
    log1p_function_name = "log1p"
    sigmoid_function_name = "sigmoid"
    softmax_function_name = "softmax"
    sqrt_function_name = "sqrt"
    tanh_function_name = "tanh"

    pow_operator = "^"

    with_sigmoid_expr = False
    with_softmax_expr = False

    def __init__(self, indent=4, function_name="score", *args, **kwargs):
        self.indent = indent
        self.function_name = function_name

        super().__init__(None, *args, **kwargs)

    def interpret(self, expr):
        top_cg = self.create_code_generator()

        self.enqueue_subroutine(self.function_name, expr)
        self.process_subroutine_queue(top_cg)

        current_dir = Path(__file__).absolute().parent

        if self.with_softmax_expr:
            filename = current_dir / "softmax.r"
            top_cg.prepend_code_lines(get_file_content(filename))

        if self.with_sigmoid_expr:
            filename = current_dir / "sigmoid.r"
            top_cg.prepend_code_lines(get_file_content(filename))

        return top_cg.finalize_and_get_generated_code()

    def create_code_generator(self):
        return RCodeGenerator(indent=self.indent)

    def interpret_softmax_expr(self, expr, **kwargs):
        self.with_softmax_expr = True
        return super().interpret_softmax_expr(expr, **kwargs)

    def interpret_sigmoid_expr(self, expr, **kwargs):
        self.with_sigmoid_expr = True
        return super().interpret_sigmoid_expr(expr, **kwargs)

    def interpret_bin_vector_expr(self, expr, **kwargs):
        self.with_linear_algebra = True
        return self._cg.infix_expression(
            left=self._do_interpret(expr.left, **kwargs),
            op=expr.op.value,
            right=self._do_interpret(expr.right, **kwargs))

    def interpret_bin_vector_num_expr(self, expr, **kwargs):
        self.with_linear_algebra = True
        return self._cg.infix_expression(
            left=self._do_interpret(expr.left, **kwargs),
            op=expr.op.value,
            right=self._do_interpret(expr.right, **kwargs))



from m2cgen.interpreters.c.interpreter import CInterpreter
from m2cgen.interpreters.c_sharp.interpreter import CSharpInterpreter
from m2cgen.interpreters.dart.interpreter import DartInterpreter
from m2cgen.interpreters.elixir.interpreter import ElixirInterpreter
from m2cgen.interpreters.f_sharp.interpreter import FSharpInterpreter
from m2cgen.interpreters.go.interpreter import GoInterpreter
from m2cgen.interpreters.haskell.interpreter import HaskellInterpreter
from m2cgen.interpreters.java.interpreter import JavaInterpreter
from m2cgen.interpreters.javascript.interpreter import JavascriptInterpreter
from m2cgen.interpreters.php.interpreter import PhpInterpreter
from m2cgen.interpreters.powershell.interpreter import PowershellInterpreter
from m2cgen.interpreters.python.interpreter import PythonInterpreter
from m2cgen.interpreters.r.interpreter import RInterpreter
from m2cgen.interpreters.ruby.interpreter import RubyInterpreter
from m2cgen.interpreters.rust.interpreter import RustInterpreter
from m2cgen.interpreters.visual_basic.interpreter import VisualBasicInterpreter

__all__ = [
    JavaInterpreter,
    PythonInterpreter,
    CInterpreter,
    GoInterpreter,
    JavascriptInterpreter,
    VisualBasicInterpreter,
    CSharpInterpreter,
    PowershellInterpreter,
    RInterpreter,
    PhpInterpreter,
    DartInterpreter,
    HaskellInterpreter,
    RubyInterpreter,
    FSharpInterpreter,
    RustInterpreter,
    ElixirInterpreter,
]



from contextlib import contextmanager

from m2cgen.ast import CompOpType
from m2cgen.interpreters.code_generator import CLikeCodeGenerator


class JavascriptCodeGenerator(CLikeCodeGenerator):

    def add_function_def(self, name, args):
        function_def = f"function {name}({', '.join(args)}) {{"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield
        self.add_block_termination()

    def vector_init(self, values):
        return f"[{', '.join(values)}]"

    def _get_var_declare_type(self, is_vector):
        return "var"

    def _comp_op_overwrite(self, op):
        if op == CompOpType.EQ:
            return "==="
        elif op == CompOpType.NOT_EQ:
            return "!=="
        else:
            return op.value
